from time import sleep
from plugin_template.data.schema import Plugin, StartPlugin, Account


RUN_USER_ACTION_NEEDED = 'userActionNeeded'
RUN_USER_ACTION_COMPLETED = 'ready'
RUN_INITIALIZED = 'initilized'
RUN_IDLE = 'idle'
RUN_STARTED = 'start'
RUN_FAILED = 'error'
RUN_COMPLETED = 'done'

AUTH_TYPES = ['userpass', 'token', 'oauth']


# A draft class to handle plugin flows like auth, state, progress...
# I think PluginFlow should be embeded in PluginBase.
class PluginFlow:

    def __init__(self, client, run_id=None, plugin=None, app_id=None, app_secret=None):
        self.client = client
        self.run_id = run_id
        self.plugin = plugin
        self.plugin_id = plugin.id if plugin else None
        self.app_id = app_id
        self.app_secret = app_secret
        self._setup_schema()
        self.initialized()

    # =======================================
    # ---------- PLUGIN METHODS -------------

    def get_account_from_plugin(self, service):
        account_edges = self.plugin.get_edges('account')
        for account_edge in account_edges:
            print("ACCOUNTS OF PLUGIN", account_edge)
        for account_edge in account_edges:
            account = account_edge.traverse(self.plugin)
            if account.service == service:
                print("Account to use", account)
                return account
        return None

    def ask_user_for_accounts(self, service, view, oauth_url=None):
        # start userActionNeeded flow
        # TODO: Handle timeouts
        vars = {
            'state': RUN_USER_ACTION_NEEDED,
            'oAuthUrl': oauth_url
        }
        self._set_run_vars(vars)
        self._set_run_view(view)

        # poll here
        while True:
            sleep(0.6)
            run_state = self._get_run_state()

            if run_state == RUN_USER_ACTION_COMPLETED:
                # Now the client has set up the account as an edge to the plugin
                self.plugin = self.client.get(self.plugin_id)
                return self.get_account_from_plugin(service)

    def initialized(self):
        self.state = RUN_INITIALIZED
        if self.run_id and self.plugin:
            self._set_run_vars({'state':RUN_INITIALIZED})

    def started(self):
        self.state = RUN_STARTED
        if self.run_id and self.plugin:
            self._set_run_vars({'state':RUN_STARTED})

    def failed(self, error):
        print("Exception while running plugin:", error)
        self.state = RUN_FAILED
        if self.run_id and self.plugin:
            self._set_run_vars({'state':RUN_FAILED, 'message': str(error)})

    def completed(self):
        self.state = RUN_COMPLETED
        if self.run_id and self.plugin:
            self._set_run_vars({'state':RUN_COMPLETED})

    # =======================================
    # --------- USER CLIENT METHODS ---------
    def install_plugin(self, name, container):
        plugin = Plugin(name=name, container=container)
        self.client.create(plugin)
        # set in instance
        self.plugin_id = plugin.id
        self.plugin = plugin
        # TODO: Add an edge from the plugin to the desired CVUStoredDefinitions
        # start_plugin.view = createLoginCVUs()
        return plugin.id

    def trigger_plugin(self):
        starter = StartPlugin(targetItemId=self.plugin_id, container=self.plugin.container, state=RUN_IDLE)
        # add cvus here
        self.client.create(starter)
        self.run_id = starter.id
        print(f"Started plugin {self.plugin.name} - {self.plugin.id} and run id {self.run_id}")

    def set_account(self, identifier, secret, service, code=None):
        # Create account item
        account = Account(identifier=identifier, secret=secret, service=service, code=code)
        # Save accounts as an edge to the plugin
        self.client.create(account)
        print('ACCOUNT', account.__dict__)
        self.plugin.add_edge('account', account)
        self.plugin.update(self.client)

    def get_CVUs(self, run):
        run = run.expand(self.client)
        return run.get_edges('view')

    # Dummy UI Client Simulation
    def run_plugin(self):
        
        self.trigger_plugin()

        # START POLLING
        while True:
            # take a breath
            sleep(0.6)
            run = self._get_run()

            if run.state == RUN_COMPLETED:
                print("# Completed plugin run. Well done")
                return True

            if run.state == RUN_USER_ACTION_NEEDED:
                # Extract views (CVUs) from the run. It should be added by the plugin.
                print(run.state)
                views = self.get_CVUs(run)
                if views:
                    for view in views:
                        print(view.__dict__)

                username = input("Username: ")
                password = input("Password: ")
                oauth_code = input("OAuth Code: ")

                ## DISPLAYS CVU & AUTHENTICATES THE USER
                sleep(3)

                self.set_account(username, password, 'demo-service', code=oauth_code)
                # Notify plugin
                self._set_run_vars({'state': RUN_USER_ACTION_COMPLETED})


    # =======================================
    # --------- INTERNAL METHODS ------------
    
    def _get_run(self):
        return self.client.get(self.run_id, expanded=False)

    def _get_run_state(self):
        start_plugin = self._get_run()
        return start_plugin.state

    def _set_run_vars(self, vars):
        start_plugin = self.client.get(self.run_id, expanded=False)
        for k,v in vars.items():
            if v:
                setattr(start_plugin, k, v)
        self.client.update_item(start_plugin)

    def _set_run_view(self, view_name):
        found_cvu = None
        views = self.client.search({'type': 'CVUStoredDefinition'}) # 'name': view_name
        for v in views:
            if v.name == view_name:
                found_cvu = v
        if not found_cvu:
            print("CVU NOT FOUND, Could not set CVU for user display")
            # raise Exception("CVU NOT FOUND")
            return
        run = self._get_run()
        run.add_edge('view', found_cvu)
        run.update()
    

    def _setup_schema(self):
        self.client.add_to_schema(Account(identifier="", secret="", service="", code=""))
        self.client.add_to_schema(Plugin(name="", container=""))
        self.client.add_to_schema(StartPlugin(targetItemId="", container="", state="", oAuthUrl=""))